# Unit test TimeSeriesFold
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from pandas.tseries.offsets import DateOffset
from skforecast.model_selection._split import TimeSeriesFold


def test_TimeSeriesFold_split_TypeError_when_X_is_not_series_dataframe_or_dict():
    """
    Test TypeError is raised when X is not a pd.Series, pd.DataFrame or dict.
    """
    X = np.arange(100)
    cv = TimeSeriesFold(steps=10, initial_train_size=70)
    err_msg = re.escape(
        f"X must be a pandas Series, DataFrame, Index or a dictionary. Got {type(X)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        cv.split(X=X)


@pytest.mark.parametrize("offset, freq", 
                         [({'days': 6}, 'D'), 
                          ({'months': 6}, 'MS')])
def test_TimeSeriesFold_split_ValueError_length_X_less_than_window_size_offset_DateOffset(offset, freq):
    """
    Test ValueError is raised when length of y is less than window_size
    when offset is a pandas DateOffset.
    """

    y = pd.Series(
        np.arange(5),
        index=pd.date_range(start='01/01/2021', periods=5, freq=freq)
    )
    
    cv = TimeSeriesFold(
             steps=2, 
             initial_train_size=3, 
             window_size=DateOffset(**offset)
         )

    err_msg = re.escape(
        f"The length of `y` (5), must be greater than or equal "
        f"to the window size ({cv.window_size}). This is because  "
        f"the offset (forecaster.offset) is larger than the available "
        f"data. Try to decrease the size of the offset (forecaster.offset), "
        f"the number of `n_offsets` (forecaster.n_offsets) or increase the "
        f"size of `y`."
    )
    with pytest.raises(ValueError, match=err_msg):
        cv.split(X=y)


def test_TimeSeriesFold_split_ValueError_when_initial_train_size_and_window_size_None():
    """
    Test ValueError is raised when initial_train_size and window_size are None.
    """
    X = pd.Series(np.arange(100))
    cv = TimeSeriesFold(steps=10, initial_train_size=None, window_size=None)
    err_msg = re.escape(
        "To use split method when `initial_train_size` is None, "
        "`window_size` must be an integer greater than 0. "
        "Although no initial training is done and all data is used to "
        "evaluate the model, the first `window_size` observations are "
        "needed to create the initial predictors. Got `window_size` = None."
    )
    with pytest.raises(ValueError, match=err_msg):
        cv.split(X=X)


def test_TimeSeriesFold_split_ValueError_when_initial_train_size_None_and_refit():
    """
    Test ValueError is raised when initial_train_size is None and refit is True.
    """
    X = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
        steps=10, initial_train_size=None, window_size=5, refit=True
    )
    err_msg = re.escape(
        "`refit` is only allowed when `initial_train_size` is not `None`. "
        "Set `refit` to `False` if you want to use `initial_train_size = None`."
    )
    with pytest.raises(ValueError, match=err_msg):
        cv.split(X=X)


def test_TimeSeriesFold_split_warning_when_window_size_is_None():
    """
    Test warning is raised when window_size is None with initial_train_size not None.
    """
    X = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
        steps=10, initial_train_size=10, window_size=None
    )
    warn_msg = re.escape(
        "Last window cannot be calculated because `window_size` is None."
    )
    with pytest.warns(UserWarning, match=warn_msg):
        cv.split(X=X)


@pytest.mark.parametrize("initial_train_size", [
    "2021-12-31",  # Before the first date in the index
    "2022-04-11",  # After the last date in the index
])
def test_TimeSeriesFold_split_invalid_initial_train_size_date(initial_train_size):
    """
    Test that ValueError is raised when initial_train_size date is outside the index range.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start="2022-01-01", periods=100, freq="D")
    cv = TimeSeriesFold(
        steps                 = 7,
        initial_train_size    = initial_train_size,
        window_size           = 10,
    )
    
    err_msg = re.escape(
        "If `initial_train_size` is a date, it must be greater than "
        "the first date in the index and less than the last date."
    )
    with pytest.raises(ValueError, match=err_msg):
        cv.split(X=y)


@pytest.mark.parametrize("initial_train_size",
                         [30, "2000-01-30 23:59:00"], 
                         ids = lambda initial_train_size: f'initial_train_size: {initial_train_size}')
def test_TimeSeriesFold_split_ValueError_window_size_as_date_offset_greater_initial_train_size(initial_train_size):
    """
    Test ValueError is raised when the window_size as date offset is greater than
    initial_train_size.
    """
    X = pd.Series(
            data  = np.arange(50),
            index = pd.date_range(start='2000-01-01', periods=50, freq='D'),
            name  = 'y'
        )
    cv = TimeSeriesFold(
        steps=5, initial_train_size=initial_train_size, window_size=pd.DateOffset(months=1)
    )
    msg = re.escape(
        "If `initial_train_size` is an integer, it must be greater than "
        "the `window_size` of the forecaster (31) "
        "and smaller than the length of the series (50). If "
        "it is a date, it must be within this range of the index."
    )
    with pytest.raises(ValueError, match=msg):
        cv.split(X=X)


def test_TimeSeriesFold_split_ValueError_when_time_series_not_enough_data_allow_incomplete_fold():
    """
    Test ValueError is raised when time series has not enough data to create 
    the folds when allow_incomplete_fold is True.
    """
    X = pd.Series(np.arange(5))
    cv = TimeSeriesFold(
        steps=10, initial_train_size=10, window_size=5, refit=True, allow_incomplete_fold=True
    )
    msg = re.escape(
        f"The time series must have more than `initial_train_size + gap` "
        f"observations to create at least one fold.\n"
        f"    Time series length: {len(X)}\n"
        f"    Required > {cv.initial_train_size + cv.gap}\n"
        f"    initial_train_size: {cv.initial_train_size}\n"
        f"    gap: {cv.gap}\n"
    )
    with pytest.raises(ValueError, match=msg):
        cv.split(X=X)


def test_TimeSeriesFold_split_ValueError_when_time_series_not_enough_data_allow_incomplete_fold_False():
    """
    Test ValueError is raised when time series has not enough data to create 
    the folds when allow_incomplete_fold is False.
    """
    X = pd.Series(np.arange(5))
    cv = TimeSeriesFold(
        steps=10, initial_train_size=10, window_size=5, refit=True, allow_incomplete_fold=False
    )
    msg = re.escape(
        f"The time series must have at least `initial_train_size + gap + steps` "
        f"observations to create a minimum of one complete fold "
        f"(allow_incomplete_fold=False).\n"
        f"    Time series length: {len(X)}\n"
        f"    Required >= {cv.initial_train_size + cv.gap + cv.steps}\n"
        f"    initial_train_size: {cv.initial_train_size}\n"
        f"    gap: {cv.gap}\n"
        f"    steps: {cv.steps}\n"
    )
    with pytest.raises(ValueError, match=msg):
        cv.split(X=X)


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(67, 70), range(70, 80), range(70, 80), True],
                                  [1, range(0, 70), range(77, 80), range(80, 90), range(80, 90), False],
                                  [2, range(0, 70), range(87, 90), range(90, 100), range(90, 100), False]]),
                          (False, [[0, [0, 70], [67, 70], [70, 80], [70, 80], True],
                                   [1, [0, 70], [77, 80], [80, 90], [80, 90], False],
                                   [2, [0, 70], [87, 90], [90, 100], [90, 100], False]])])
def test_TimeSeriesFold_split_no_refit_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is False, gap=0 and not 
    remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = 3,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 65), range(61, 65), range(65, 75), range(65, 75), True],
                                  [1, range(0, 65), range(71, 75), range(75, 85), range(75, 85), False],
                                  [2, range(0, 65), range(81, 85), range(85, 95), range(85, 95), False]]),
                          (False, [[0, [0, 65], [61, 65], [65, 75], [65, 75], True],
                                   [1, [0, 65], [71, 75], [75, 85], [75, 85], False],
                                   [2, [0, 65], [81, 85], [85, 95], [85, 95], False]])])
def test_TimeSeriesFold_split_no_refit_no_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 0 (False), gap=0, 
    remainder and allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 65,
            window_size           = 4,
            differentiation       = None,
            refit                 = 0,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 65\n"
        "Number of observations used for backtesting: 35\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n"
        "    The last 1 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-06 00:00:00  (n=65)\n"
        "    Validation: 2022-03-07 00:00:00 -- 2022-03-16 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-26 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-27 00:00:00 -- 2022-04-05 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(65, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 70), range(72, 77), range(77, 89), range(82, 89), False],
                                  [2, range(0, 70), range(79, 84), range(84, 96), range(89, 96), False],
                                  [3, range(0, 70), range(86, 91), range(91, 100), range(96, 100), False]]),
                          (False, [[0, [0, 70], [65, 70], [70, 82], [75, 82], True],
                                   [1, [0, 70], [72, 77], [77, 89], [82, 89], False],
                                   [2, [0, 70], [79, 84], [84, 96], [89, 96], False],
                                   [3, [0, 70], [86, 91], [91, 100], [96, 100], False]])])
def test_TimeSeriesFold_split_no_refit_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is False, gap=5, 
    remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 5,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 15), range(0, 15), range(15, 30), range(20, 30), False],
                                  [1, range(0, 15), range(10, 25), range(25, 40), range(30, 40), False],
                                  [2, range(0, 15), range(20, 35), range(35, 50), range(40, 50), False],
                                  [3, range(0, 15), range(30, 45), range(45, 60), range(50, 60), False],
                                  [4, range(0, 15), range(40, 55), range(55, 70), range(60, 70), False],
                                  [5, range(0, 15), range(50, 65), range(65, 80), range(70, 80), False],
                                  [6, range(0, 15), range(60, 75), range(75, 90), range(80, 90), False],
                                  [7, range(0, 15), range(70, 85), range(85, 100), range(90, 100), False]]),
                          (False, [[0, [0, 15], [0, 15], [15, 30], [20, 30], False],
                                   [1, [0, 15], [10, 25], [25, 40], [30, 40], False],
                                   [2, [0, 15], [20, 35], [35, 50], [40, 50], False],
                                   [3, [0, 15], [30, 45], [45, 60], [50, 60], False],
                                   [4, [0, 15], [40, 55], [55, 70], [60, 70], False],
                                   [5, [0, 15], [50, 65], [65, 80], [70, 80], False],
                                   [6, [0, 15], [60, 75], [75, 90], [80, 90], False],
                                   [7, [0, 15], [70, 85], [85, 100], [90, 100], False]])])
def test_TimeSeriesFold_split_no_refit_initial_train_size_None_gap(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is False, gap=5, 
    initial_train_size=None (externally fitted), no remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
             steps                 = 10,
             initial_train_size    = None,
             window_size           = 15,
             differentiation       = None,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 5,
             skip_folds            = None,
             allow_incomplete_fold = True,
             return_all_indexes    = return_all_indexes,
         )
    folds = cv.split(X=y)            
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "An already trained forecaster is to be used. Window size: 15\n"
        "Number of observations used for backtesting: 85\n"
        "    Number of folds: 8\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n\n"
        "Fold: 0\n"
        "    Training:   No training in this fold\n"
        "    Validation: 20 -- 29  (n=10)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 30 -- 39  (n=10)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 40 -- 49  (n=10)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 50 -- 59  (n=10)\n"
        "Fold: 4\n"
        "    Training:   No training in this fold\n"
        "    Validation: 60 -- 69  (n=10)\n"
        "Fold: 5\n"
        "    Training:   No training in this fold\n"
        "    Validation: 70 -- 79  (n=10)\n"
        "Fold: 6\n"
        "    Training:   No training in this fold\n"
        "    Validation: 80 -- 89  (n=10)\n"
        "Fold: 7\n"
        "    Training:   No training in this fold\n"
        "    Validation: 90 -- 99  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(68, 70), range(70, 80), range(70, 80), True],
                                  [1, range(0, 80), range(78, 80), range(80, 90), range(80, 90), True],
                                  [2, range(0, 90), range(88, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[0, [0, 70], [68, 70], [70, 80], [70, 80], True],
                                   [1, [0, 80], [78, 80], [80, 90], [80, 90], True],
                                   [2, [0, 90], [88, 90], [90, 100], [90, 100], True]])])
def test_TimeSeriesFold_split_refit_no_fixed_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    False, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = 2,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )

    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-21 00:00:00  (n=80)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-31 00:00:00  (n=90)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(64, 70), range(70, 80), range(70, 80), True],
                                  [1, range(10, 80), range(74, 80), range(80, 90), range(80, 90), True],
                                  [2, range(20, 90), range(84, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[0, [0, 70], [64, 70], [70, 80], [70, 80], True],
                                   [1, [10, 80], [74, 80], [80, 90], [80, 90], True],
                                   [2, [20, 90], [84, 90], [90, 100], [90, 100], True]])])
def test_TimeSeriesFold_split_refit_fixed_train_size_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 1 (True), fixed_train_size 
    is True, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = 6,
            differentiation       = None,
            refit                 = 1,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-11 00:00:00 -- 2022-03-21 00:00:00  (n=70)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-21 00:00:00 -- 2022-03-31 00:00:00  (n=70)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(67, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 77), range(74, 77), range(77, 89), range(82, 89), True],
                                  [2, range(0, 84), range(81, 84), range(84, 96), range(89, 96), True],
                                  [3, range(0, 91), range(88, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[0, [0, 70], [67, 70], [70, 82], [75, 82], True],
                                   [1, [0, 77], [74, 77], [77, 89], [82, 89], True],
                                   [2, [0, 84], [81, 84], [84, 96], [89, 96], True],
                                   [3, [0, 91], [88, 91], [91, 100], [96, 100], True]])])
def test_TimeSeriesFold_split_refit_no_fixed_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 3,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-18 00:00:00  (n=77)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-04-01 00:00:00  (n=91)\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(65, 70), range(70, 82), range(75, 82), True],
                                  [1, range(7, 77), range(72, 77), range(77, 89), range(82, 89), True],
                                  [2, range(14, 84), range(79, 84), range(84, 96), range(89, 96), True],
                                  [3, range(21, 91), range(86, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[0, [0, 70], [65, 70], [70, 82], [75, 82], True],
                                   [1, [7, 77], [72, 77], [77, 89], [82, 89], True],
                                   [2, [14, 84], [79, 84], [84, 96], [89, 96], True],
                                   [3, [21, 91], [86, 91], [91, 100], [96, 100], True]])])
def test_TimeSeriesFold_split_refit_fixed_train_size_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 5,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   7 -- 76  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n"
        "Fold: 3\n"
        "    Training:   21 -- 90  (n=70)\n"
        "    Validation: 96 -- 99  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(67, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 77), range(74, 77), range(77, 89), range(82, 89), True],
                                  [2, range(0, 84), range(81, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[0, [0, 70], [67, 70], [70, 82], [75, 82], True],
                                   [1, [0, 77], [74, 77], [77, 89], [82, 89], True],
                                   [2, [0, 84], [81, 84], [84, 96], [89, 96], True]])])
def test_TimeSeriesFold_split_refit_no_fixed_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 3,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    The last 1 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-18 00:00:00  (n=77)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(66, 70), range(70, 82), range(75, 82), True],
                                  [1, range(7, 77), range(73, 77), range(77, 89), range(82, 89), True],
                                  [2, range(14, 84), range(80, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[0, [0, 70], [66, 70], [70, 82], [75, 82], True],
                                   [1, [7, 77], [73, 77], [77, 89], [82, 89], True],
                                   [2, [14, 84], [80, 84], [84, 96], [89, 96], True]])])
def test_TimeSeriesFold_split_refit_fixed_train_size_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 4,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    The last 1 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   7 -- 76  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 60), range(56, 60), range(60, 70), range(60, 70), True],
                                  [1, range(0, 60), range(66, 70), range(70, 80), range(70, 80), False],
                                  [2, range(0, 80), range(76, 80), range(80, 90), range(80, 90), True],
                                  [3, range(0, 80), range(86, 90), range(90, 100), range(90, 100), False]]),
                          (False, [[0, [0, 60], [56, 60], [60, 70], [60, 70], True],
                                   [1, [0, 60], [66, 70], [70, 80], [70, 80], False],
                                   [2, [0, 80], [76, 80], [80, 90], [80, 90], True],
                                   [3, [0, 80], [86, 90], [90, 100], [90, 100], False]])])
def test_TimeSeriesFold_split_refit_int_no_fixed_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 2, fixed_train_size is 
    False, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 60,
            window_size           = 4,
            differentiation       = None,
            refit                 = 2,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 60\n"
        "Number of observations used for backtesting: 40\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-02 00:00:00 -- 2022-03-11 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-21 00:00:00  (n=80)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 60), range(58, 60), range(60, 70), range(60, 70), True],
                                  [1, range(0, 60), range(68, 70), range(70, 80), range(70, 80), False],
                                  [2, range(0, 60), range(78, 80), range(80, 90), range(80, 90), False],
                                  [3, range(30, 90), range(88, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[0, [0, 60], [58, 60], [60, 70], [60, 70], True],
                                   [1, [0, 60], [68, 70], [70, 80], [70, 80], False],
                                   [2, [0, 60], [78, 80], [80, 90], [80, 90], False],
                                   [3, [30, 90], [88, 90], [90, 100], [90, 100], True]])])
def test_TimeSeriesFold_split_refit_int_fixed_train_size_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 3, fixed_train_size is 
    True, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 60,
            window_size           = 2,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 60\n"
        "Number of observations used for backtesting: 40\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-02 00:00:00 -- 2022-03-11 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 3\n"
        "    Training:   2022-01-31 00:00:00 -- 2022-03-31 00:00:00  (n=60)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(60, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 70), range(67, 77), range(77, 89), range(82, 89), False],
                                  [2, range(0, 84), range(74, 84), range(84, 96), range(89, 96), True],
                                  [3, range(0, 84), range(81, 91), range(91, 100), range(96, 100), False]]),
                          (False, [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                                   [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                                   [2, [0, 84], [74, 84], [84, 96], [89, 96], True],
                                   [3, [0, 84], [81, 91], [91, 100], [96, 100], False]])])
def test_TimeSeriesFold_split_refit_int_no_fixed_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 2, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 10,
            differentiation       = None,
            refit                 = 2,
            fixed_train_size      = False,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(55, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 70), range(62, 77), range(77, 89), range(82, 89), False],
                                  [2, range(0, 70), range(69, 84), range(84, 96), range(89, 96), False],
                                  [3, range(21, 91), range(76, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[0, [0, 70], [55, 70], [70, 82], [75, 82], True],
                                   [1, [0, 70], [62, 77], [77, 89], [82, 89], False],
                                   [2, [0, 70], [69, 84], [84, 96], [89, 96], False],
                                   [3, [21, 91], [76, 91], [91, 100], [96, 100], True]])])
def test_TimeSeriesFold_split_refit_int_fixed_train_size_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 3, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 15,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = True,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 89 -- 95  (n=7)\n"
        "Fold: 3\n"
        "    Training:   21 -- 90  (n=70)\n"
        "    Validation: 96 -- 99  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(50, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 70), range(57, 77), range(77, 89), range(82, 89), False],
                                  [2, range(0, 70), range(64, 84), range(84, 96), range(89, 96), False]]),
                          (False, [[0, [0, 70], [50, 70], [70, 82], [75, 82], True],
                                   [1, [0, 70], [57, 77], [77, 89], [82, 89], False],
                                   [2, [0, 70], [64, 84], [84, 96], [89, 96], False]])])
def test_TimeSeriesFold_split_refit_int_no_fixed_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 3, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 20,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = False,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    The last 1 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(60, 70), range(70, 82), range(75, 82), True],
                                  [1, range(0, 70), range(67, 77), range(77, 89), range(82, 89), False],
                                  [2, range(14, 84), range(74, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                                   [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                                   [2, [14, 84], [74, 84], [84, 96], [89, 96], True]])])
def test_TimeSeriesFold_split_refit_int_fixed_train_size_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = 70,
            window_size           = 10,
            differentiation       = None,
            refit                 = 2,
            fixed_train_size      = True,
            gap                   = 5,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    The last 1 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(67, 70), range(70, 80), range(70, 80), True],
                                  [1, range(0, 80), range(77, 80), range(80, 90), range(80, 90), True],
                                  [2, range(0, 90), range(87, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[0, [0, 70], [67, 70], [70, 80], [70, 80], True],
                                   [1, [0, 80], [77, 80], [80, 90], [80, 90], True],
                                   [2, [0, 90], [87, 90], [90, 100], [90, 100], True]])])
def test_TimeSeriesFold_split_refit_no_fixed_no_gap_no_remainder_differentiation(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    False, gap=0, not remainder and differentiation=1.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    differentiation = 1
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = 2 + differentiation,
            differentiation       = differentiation,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 69\n"
        "    First 1 observation/s in training sets are used for differentiation\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-11 00:00:00  (n=69)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-21 00:00:00  (n=79)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-31 00:00:00  (n=89)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("skip_folds",
                         [3, [1, 2, 4, 5, 7]],
                         ids = lambda skip_folds: f'{skip_folds}')
@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(68, 70), range(70, 80), range(70, 80), True],
                                  [3, range(30, 100), range(98, 100), range(100, 110), range(100, 110), True],
                                  [6, range(60, 130), range(128, 130), range(130, 140), range(130, 140), True]]),
                          (False, [[0, [0, 70], [68, 70], [70, 80], [70, 80], True],
                                   [3, [30, 100], [98, 100], [100, 110], [100, 110], True],
                                   [6, [60, 130], [128, 130], [130, 140], [130, 140], True]])])
def test_TimeSeriesFold_split_refit_fixed_no_gap_no_remainder_skip_folds_3(capfd, skip_folds, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, fixed_train_size is 
    True, gap=0, not remainder and skip_folds=3.
    """
    y = pd.Series(np.arange(150))
    y.index = pd.date_range(start='2022-01-01', periods=150, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = 2,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = skip_folds,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)

    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 80\n"
        "    Number of folds: 8\n"
        "    Number skipped folds: 5 [1, 2, 4, 5, 7]\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Fold skipped\n"
        "Fold: 2\n"
        "    Fold skipped\n"
        "Fold: 3\n"
        "    Training:   2022-01-31 00:00:00 -- 2022-04-10 00:00:00  (n=70)\n"
        "    Validation: 2022-04-11 00:00:00 -- 2022-04-20 00:00:00  (n=10)\n"
        "Fold: 4\n"
        "    Fold skipped\n"
        "Fold: 5\n"
        "    Fold skipped\n"
        "Fold: 6\n"
        "    Training:   2022-03-02 00:00:00 -- 2022-05-10 00:00:00  (n=70)\n"
        "    Validation: 2022-05-11 00:00:00 -- 2022-05-20 00:00:00  (n=10)\n"
        "Fold: 7\n"
        "    Fold skipped\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("window_size",
                         [3, None], 
                         ids = lambda ws: f'window_size: {ws}')
def test_TimeSeriesFold_split_as_pandas_return_all_indexes_True(window_size):
    """
    Test TimeSeriesFold split method output when as_pandas=True and return_all_indexes=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = window_size,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            verbose               = False,
            return_all_indexes    = True,
        )
    folds = cv.split(X=y, as_pandas=True)

    expected = pd.DataFrame(
        {'fold': [0, 1, 2],
         'train_index': [range(0, 70), range(0, 70), range(0, 70)],
         'last_window_index': [range(67, 70), range(77, 80), range(87, 90)],
         'test_index': [range(70, 80), range(80, 90), range(90, 100)],
         'test_index_with_gap': [range(70, 80), range(80, 90), range(90, 100)],
         'fit_forecaster': [True, False, False]}
    )

    if window_size is None:
        expected['last_window_index'] = [[None, None], [None, None], [None, None]]
    
    pd.testing.assert_frame_equal(folds, expected)


@pytest.mark.parametrize("window_size",
                         [3, None], 
                         ids = lambda ws: f'window_size: {ws}')
def test_TimeSeriesFold_split_as_pandas_return_all_indexes_False(window_size):
    """
    Test TimeSeriesFold split method output when as_pandas=True and return_all_indexes=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = 70,
            window_size           = window_size,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = [1],
            allow_incomplete_fold = True,
            verbose               = False,
            return_all_indexes    = False
        )
    folds = cv.split(X=y, as_pandas=True)

    expected = pd.DataFrame(
        {'fold': [0, 2],
         'train_start': [0, 0],
         'train_end': [70, 70],
         'last_window_start': [67, 87],
         'last_window_end': [70, 90],
         'test_start': [70, 90],
         'test_end': [80, 100],
         'test_start_with_gap': [70, 90],
         'test_end_with_gap': [80, 100],
         'fit_forecaster': [True, False]}
    )

    if window_size is None:
        expected['last_window_start'] = [None, None]
        expected['last_window_end'] = [None, None]
    
    pd.testing.assert_frame_equal(folds, expected)


@pytest.mark.parametrize("initial_train_size, expected",
                         [(70, 
                           [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                            [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                            [2, [0, 70], [74, 84], [84, 96], [89, 96], False],
                            [3, [0, 70], [81, 91], [91, 100], [96, 100], False]]),
                            ("2022-03-11", 
                           [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                            [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                            [2, [0, 70], [74, 84], [84, 96], [89, 96], False],
                            [3, [0, 70], [81, 91], [91, 100], [96, 100], False]]),
                            ("2022-03-11 00:00:00",
                           [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                            [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                            [2, [0, 70], [74, 84], [84, 96], [89, 96], False],
                            [3, [0, 70], [81, 91], [91, 100], [96, 100], False]]),
                            (pd.to_datetime("2022-03-11"),
                           [[0, [0, 70], [60, 70], [70, 82], [75, 82], True],
                            [1, [0, 70], [67, 77], [77, 89], [82, 89], False],
                            [2, [0, 70], [74, 84], [84, 96], [89, 96], False],
                            [3, [0, 70], [81, 91], [91, 100], [96, 100], False]])])
def test_TimeSeriesFold_split_int_and_date_initial_train_size(capfd, initial_train_size, expected):
    """
    Test TimeSeriesFold split method output when initial_train_size is 
    an integer or a date in string and pandas datetime format.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 7,
            initial_train_size    = initial_train_size,
            window_size           = 10,
            gap                   = 5,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(66, 70), range(70, 81), range(70, 81), True],
                                  [1, range(0, 70), range(71, 75), range(75, 86), range(75, 86), False],
                                  [2, range(0, 70), range(76, 80), range(80, 91), range(80, 91), False],
                                  [3, range(0, 70), range(81, 85), range(85, 96), range(85, 96), False]]),
                          (False, [[0, [0, 70], [66, 70], [70, 81], [70, 81], True],
                                   [1, [0, 70], [71, 75], [75, 86], [75, 86], False],
                                   [2, [0, 70], [76, 80], [80, 91], [80, 91], False],
                                   [3, [0, 70], [81, 85], [85, 96], [85, 96], False]])])
def test_TimeSeriesFold_split_no_refit_no_gap_allow_incomplete_fold_False_fold_stride(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is 0 (False), gap=0, 
    remainder, allow_incomplete_fold=False and fold_stride=5.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = TimeSeriesFold(
            steps                 = 11,
            initial_train_size    = 70,
            fold_stride           = 5,
            window_size           = 4,
            differentiation       = None,
            refit                 = 0,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 11\n"
        "    Number of steps to the next fold (fold stride): 5\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 0\n"
        "    The last 2 fold(s) have been excluded because they were incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-22 00:00:00  (n=11)\n"
        "Fold: 1\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-27 00:00:00  (n=11)\n"
        "Fold: 2\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-04-01 00:00:00  (n=11)\n"
        "Fold: 3\n"
        "    Training:   No training in this fold\n"
        "    Validation: 2022-03-27 00:00:00 -- 2022-04-06 00:00:00  (n=11)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[0, range(0, 70), range(64, 70), range(70, 83), range(72, 83), True],
                                  [1, range(0, 75), range(69, 75), range(75, 88), range(77, 88), True],
                                  [2, range(0, 80), range(74, 80), range(80, 93), range(82, 93), True],
                                  [3, range(0, 85), range(79, 85), range(85, 98), range(87, 98), True],
                                  [4, range(0, 90), range(84, 90), range(90, 100), range(92, 100), True],
                                  [5, range(0, 95), range(89, 95), range(95, 100), range(97, 100), True]]),
                          (False, [[0, [0, 70], [64, 70], [70, 83], [72, 83], True],
                                   [1, [0, 75], [69, 75], [75, 88], [77, 88], True],
                                   [2, [0, 80], [74, 80], [80, 93], [82, 93], True],
                                   [3, [0, 85], [79, 85], [85, 98], [87, 98], True],
                                   [4, [0, 90], [84, 90], [90, 100], [92, 100], True],
                                   [5, [0, 95], [89, 95], [95, 100], [97, 100], True]])])
def test_TimeSeriesFold_split_no_refit_gap_allow_incomplete_fold_True_fold_stride(capfd, return_all_indexes, expected):
    """
    Test TimeSeriesFold split method output when refit is True, gap=2, 
    remainder, allow_incomplete_fold=True and fold_stride=5.
    """
    y = pd.Series(np.arange(100))
    cv = TimeSeriesFold(
            steps                 = 11,
            initial_train_size    = 70,
            fold_stride           = 5,
            window_size           = 6,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 2,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = return_all_indexes,
        )
    folds = cv.split(X=y)
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 6\n"
        "    Number skipped folds: 0 \n"
        "    Number of steps per fold: 11\n"
        "    Number of steps to the next fold (fold stride): 5\n"
        "    Number of steps to exclude between last observed data (last window) and predictions (gap): 2\n"
        "    Last fold only includes 3 observations.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 72 -- 82  (n=11)\n"
        "Fold: 1\n"
        "    Training:   0 -- 74  (n=75)\n"
        "    Validation: 77 -- 87  (n=11)\n"
        "Fold: 2\n"
        "    Training:   0 -- 79  (n=80)\n"
        "    Validation: 82 -- 92  (n=11)\n"
        "Fold: 3\n"
        "    Training:   0 -- 84  (n=85)\n"
        "    Validation: 87 -- 97  (n=11)\n"
        "Fold: 4\n"
        "    Training:   0 -- 89  (n=90)\n"
        "    Validation: 92 -- 99  (n=8)\n"
        "Fold: 5\n"
        "    Training:   0 -- 94  (n=95)\n"
        "    Validation: 97 -- 99  (n=3)\n\n"
    )

    assert out == expected_out
    assert folds == expected



# Unit test set_out_sample_residuals ForecasterEquivalentDate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from pandas.tseries.offsets import DateOffset
from sklearn.exceptions import NotFittedError
from skforecast.exceptions import ResidualsUsageWarning
from skforecast.recursive import ForecasterEquivalentDate

# Fixtures
from .fixtures_forecaster_equivalent_date import y


def test_set_out_sample_residuals_NotFittedError_when_forecaster_not_fitted():
    """
    Test NotFittedError is raised when forecaster is not fitted.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    y_true = {1: np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}
    y_pred = {1: np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `set_out_sample_residuals()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_is_not_numpy_array_or_pandas_series():
    """
    Test TypeError is raised when y_true argument is not numpy ndarray or pandas Series.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y)
    y_true = 'invalid'
    y_pred = np.array([1, 2, 3])

    err_msg = re.escape(
        f"`y_true` argument must be `numpy ndarray` or `pandas Series`. "
        f"Got {type(y_true)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_is_not_numpy_array_or_pandas_series():
    """
    Test TypeError is raised when y_pred argument is not numpy ndarray or pandas Series.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y)
    y_true = np.array([1, 2, 3])
    y_pred = 'invalid'

    err_msg = re.escape(
        f"`y_pred` argument must be `numpy ndarray` or `pandas Series`. "
        f"Got {type(y_pred)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_different_length():
    """
    Test ValueError is raised when y_true and y_pred have different length.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y)
    y_true = np.array([1, 2, 3])
    y_pred = np.array([1, 2, 3, 4])

    err_msg = re.escape(
        f"`y_true` and `y_pred` must have the same length. "
        f"Got {len(y_true)} and {len(y_pred)}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_different_index():
    """
    Test ValueError is raised when residuals and y_pred have different index.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y)
    y_true = pd.Series([1, 2, 3], index=[1, 2, 3])
    y_pred = pd.Series([1, 2, 3], index=[1, 2, 4])

    err_msg = re.escape("`y_true` and `y_pred` must have the same index.")
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is False.
    """
    rng = np.random.default_rng(12345)
    y_true = pd.Series(rng.normal(loc=10, scale=10, size=1000))
    y_pred = pd.Series(rng.normal(loc=10, scale=10, size=1000))

    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y_true)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=False)
    results = np.sort(forecaster.out_sample_residuals_)

    expected = np.sort(y_true - y_pred)

    np.testing.assert_array_almost_equal(results, expected)


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is True.
    """
    rng = np.random.default_rng(12345)
    y_true = pd.Series(rng.normal(loc=10, scale=10, size=1000))
    y_true.index = pd.date_range(start='2000-01-01', periods=1000, freq='D')
    y_pred = pd.Series(rng.normal(loc=10, scale=10, size=1000))
    y_pred.index = pd.date_range(start='2000-01-01', periods=1000, freq='D')

    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    forecaster.fit(y_true)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
    results = np.sort(forecaster.out_sample_residuals_)
    
    residuals = (y_true - y_pred)
    expected = np.sort(np.concatenate((residuals, residuals)))

    np.testing.assert_array_almost_equal(results, expected)


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000():
    """
    Test length residuals stored when its length is greater than 10_000.
    """
    rng = np.random.RandomState(42)
    y_fit = pd.Series(rng.normal(loc=10, scale=10, size=50_000))
    y_fit.index = pd.date_range(start='2000-01-01', periods=50_000, freq='D')

    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {"n_bins": 10}
                 )
    forecaster.fit(y_fit)

    y_pred = [
        y_fit.loc[(y_fit.index - forecaster.offset * n_off)[forecaster.window_size:]]
        for n_off in range(1, forecaster.n_offsets + 1)
    ]

    y_pred = np.apply_along_axis(
                 forecaster.agg_func,
                 axis = 0,
                 arr  = np.vstack(y_pred)
             )

    forecaster.set_out_sample_residuals(
        y_true = y_fit.to_numpy()[forecaster.window_size:],
        y_pred = y_pred
    )

    assert len(forecaster.out_sample_residuals_) == 10_000
    for v in forecaster.out_sample_residuals_by_bin_.values():
        assert len(v) == 1_000


def test_out_sample_residuals_by_bin_and_in_sample_residuals_by_bin_equivalence():
    """
    Test out sample residuals by bin are equivalent to in-sample residuals by bin
    when training data and training predictions are passed.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {"n_bins": 3}
                 )
    forecaster.fit(y, store_in_sample_residuals=True)

    y_pred = [
        y.shift(forecaster.offset * n_off)[forecaster.window_size:]
        for n_off in range(1, forecaster.n_offsets + 1)
    ]

    y_pred = np.apply_along_axis(
                 forecaster.agg_func,
                 axis = 0,
                 arr  = np.vstack(y_pred)
             )

    forecaster.set_out_sample_residuals(
        y_true = y.to_numpy()[forecaster.window_size:],
        y_pred = y_pred
    )

    assert forecaster.in_sample_residuals_by_bin_.keys() == forecaster.out_sample_residuals_by_bin_.keys()
    for k in forecaster.out_sample_residuals_by_bin_.keys():
        np.testing.assert_array_almost_equal(
            forecaster.in_sample_residuals_by_bin_[k],
            forecaster.out_sample_residuals_by_bin_[k]
        )


def test_set_out_sample_residuals_append_new_residuals_per_bin():
    """
    Test that set_out_sample_residuals append residuals per bin until it
    reaches the max allowed size of 10_000 // n_bins
    """
    rng = np.random.default_rng(12345)
    y_fit = pd.Series(
                data=rng.normal(loc=10, scale=1, size=1001),
                index=pd.date_range(start="01-01-2000", periods=1001, freq="h"),
            )

    forecaster = ForecasterEquivalentDate(
                     offset        = 1,
                     n_offsets     = 1,
                     agg_func      = np.mean,
                     binner_kwargs = {"n_bins": 2}
                 )
    forecaster.fit(y_fit)

    y_pred = [
        y_fit.shift(forecaster.offset * n_off)[forecaster.window_size:]
        for n_off in range(1, forecaster.n_offsets + 1)
    ]

    y_pred = np.apply_along_axis(
                 forecaster.agg_func,
                 axis = 0,
                 arr  = np.vstack(y_pred)
             )

    for i in range(1, 20):
        forecaster.set_out_sample_residuals(
            y_true = y_fit.to_numpy()[forecaster.window_size:],
            y_pred = y_pred,
            append = True
        )
        for v in forecaster.out_sample_residuals_by_bin_.values():
            assert len(v) == min(5_000, 500 * i)


def test_set_out_sample_residuals_when_there_are_no_residuals_for_some_bins():
    """
    Test that set_out_sample_residuals works when there are no residuals for some bins.
    """
    rng = np.random.default_rng(12345)
    y = pd.Series(
            data=rng.normal(loc=10, scale=1, size=100),
            index=pd.date_range(start="01-01-2000", periods=100, freq="h"),
        )

    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {"n_bins": 3}
                 )
    forecaster.fit(y)
    y_pred = y.loc[y > 10]
    y_true = y_pred + rng.normal(loc=0, scale=1, size=len(y_pred))

    warn_msg = re.escape(
        f"The following bins have no out of sample residuals: [0]. "
        f"No predicted values fall in the interval "
        f"[{forecaster.binner_intervals_[0]}]. "
        f"Empty bins will be filled with a random sample of residuals."
    )
    with pytest.warns(ResidualsUsageWarning, match=warn_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)

    assert len(forecaster.out_sample_residuals_by_bin_[0]) == len(y_pred)



# Unit test predict_quantiles ForecasterRecursiveMultiSeries
# ==============================================================================
import pytest
import pandas as pd
from ....recursive import ForecasterRecursiveMultiSeries
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_dict_range,
    exog_wide_range,
    exog_pred_wide_range
)

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


@pytest.mark.parametrize("level", 
                         ['l1', ['l1']], 
                         ids=lambda lvl: f'level: {lvl}')
def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer(level):
    """
    Test output of predict_quantiles when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed. Single level.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  quantiles               = [0.05, 0.55, 0.95],
                  levels                  = level,
                  exog                    = exog_pred_wide_range,
                  n_boot                  = 4,
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False,
                  suppress_warnings       = True
              )
    
    expected = pd.DataFrame(
        {'level': ['l1', 'l1'],
         'q_0.05': [0.1283750019253314, 0.09385929028369558],
         'q_0.55': [0.3392034161273868, 0.3207904249631374],
         'q_0.95': [0.47157639833964976, 0.6231596160709784]},
        index = pd.RangeIndex(start=50, stop=52)
    )
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("levels", 
                         [['l1', 'l2'], None], 
                         ids=lambda lvl: f'levels: {lvl}')
def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer(levels):
    """
    Test output of predict_quantiles when estimator is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed. Multiple levels.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )

    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  quantiles               = (0.05, 0.55, 0.95),
                  levels                  = levels,
                  exog                    = exog_pred_wide_range,
                  n_boot                  = 4,
                  use_in_sample_residuals = False,
                  use_binned_residuals    = False
              )
    
    expected = pd.DataFrame(
        {
            "level": ["l1", "l2", "l1", "l2"],
            "q_0.05": [
                0.1283750019253314,
                0.1615361493231256,
                0.09385929028369558,
                0.07374959117551036,
            ],
            "q_0.55": [
                0.3392034161273868,
                0.6552169189586099,
                0.3207904249631374,
                0.10749930737109713,
            ],
            "q_0.95": [
                0.47157639833964976,
                0.9245514385384845,
                0.6231596160709784,
                0.2184633069802528,
            ],
        },
        index=pd.Index([50, 50, 51, 51]),
    )

    pd.testing.assert_frame_equal(results, expected)



# Unit test get_style_repr_html
# ==============================================================================
import pytest
from ...utils import get_style_repr_html


@pytest.mark.parametrize("is_fitted", 
                         [True, False], 
                         ids=lambda is_fitted: f'is_fitted: {is_fitted}')
def test_get_style_repr_html(is_fitted):
    """
    Check if the style is correct.
    """

    style, unique_id = get_style_repr_html(is_fitted=is_fitted)
    
    background_color = "#f0f8ff" if is_fitted else "#f9f1e2"
    section_color = "#b3dbfd" if is_fitted else "#fae3b3"

    expected_style = f"""
    <style>
        .container-{unique_id} {{
            font-family: 'Arial', sans-serif;
            font-size: 0.9em;
            color: #333333;
            border: 1px solid #ddd;
            background-color: {background_color};
            padding: 5px 15px;
            border-radius: 8px;
            max-width: 600px;
            #margin: auto;
        }}
        .container-{unique_id} h2 {{
            font-size: 1.5em;
            color: #222222;
            border-bottom: 2px solid #ddd;
            padding-bottom: 5px;
            margin-bottom: 15px;
            margin-top: 5px;
        }}
        .container-{unique_id} details {{
            margin: 10px 0;
        }}
        .container-{unique_id} summary {{
            font-weight: bold;
            font-size: 1.1em;
            color: #000000;
            cursor: pointer;
            margin-bottom: 5px;
            background-color: {section_color};
            padding: 5px;
            border-radius: 5px;
        }}
        .container-{unique_id} summary:hover {{
            color: #000000;
            background-color: #e0e0e0;
        }}
        .container-{unique_id} ul {{
            font-family: 'Courier New', monospace;
            list-style-type: none;
            padding-left: 20px;
            margin: 10px 0;
            line-height: normal;
        }}
        .container-{unique_id} li {{
            margin: 5px 0;
            font-family: 'Courier New', monospace;
        }}
        .container-{unique_id} li strong {{
            font-weight: bold;
            color: #444444;
        }}
        .container-{unique_id} li::before {{
            content: "- ";
            color: #666666;
        }}
        .container-{unique_id} a {{
            color: #001633;
            text-decoration: none;
        }}
        .container-{unique_id} a:hover {{
            color: #359ccb; 
        }}
    </style>
    """

    assert isinstance(unique_id, str)
    assert style == expected_style



# Unit test fit
# ==============================================================================
import pandas as pd
import numpy as np
import joblib
from pathlib import Path
import pytest
from ....drift_detection import PopulationDriftDetector


# fixtures
THIS_DIR = Path(__file__).parent
data = joblib.load(THIS_DIR/'fixture_data_population_drift.joblib')
data_multiseries = pd.concat(
    [
        data.assign(series='series_1'),
        data.assign(series='series_2'),
        data.assign(series='series_3')
    ]
).set_index('series', append=True).swaplevel(0, 1)


def test_fit_exception_X_is_not_dataframe():
    """
    Test that fit raises an exception when X is not a pandas DataFrame.
    """
    detector = PopulationDriftDetector(chunk_size="ME", threshold=0.95)
    X = "not a dataframe"
    err_msg = f"`X` must be a pandas DataFrame. Got {type(X)} instead."
    with pytest.raises(ValueError, match=err_msg):
        detector.fit(X=X)


def test_fit_exception_chunk_size_is_pandas_DateOffset_str_but_X_has_no_datetime_index():
    """
    Test that fit raises an exception when chunk_size is a string compatible with
    pandas DateOffset but X has no datetime index.
    """
    detector = PopulationDriftDetector(chunk_size="ME", threshold=0.95)
    X = pd.DataFrame({"A": np.random.rand(100), "B": np.random.rand(100)})
    err_msg = (
        "`chunk_size` is a pandas DateOffset but `X` does not have a DatetimeIndex."
    )
    with pytest.raises(ValueError, match=err_msg):
        detector.fit(X=X)


def test_fit_exception_warning_when_feature_is_all_nan():
    """
    Test that fit raises a warning when a feature is all NaN.
    """
    detector = PopulationDriftDetector(chunk_size="ME", threshold=0.95)
    X = data.copy()
    X.loc[:, "temp"] = np.nan
    warn_msg = (
        "Feature 'temp' contains only NaN values in the reference dataset. "
        "Drift detection skipped."
    )
    with pytest.warns(UserWarning, match=warn_msg):
        detector.fit(X=X)


def test_fit_stored_attributes():
    """
    Test that the attributes of PopulationDriftDetector are correctly stored after calling fit.
    """

    detector = PopulationDriftDetector(chunk_size="ME", threshold=0.95)
    detector.fit(data)

    expected_results = {
        "ref_features_": [
            "holiday",
            "workingday",
            "weather",
            "temp",
            "atemp",
            "hum",
            "windspeed",
            "users",
            "month",
            "hour",
            "weekday",
        ],
        "is_fitted_": True,
        "empirical_threshold_ks_": {
            "holiday": 0.03659032405160117,
            "workingday": 0.05896694761925775,
            "weather": np.nan,
            "temp": 0.6644120588970772,
            "atemp": 0.6419687642498859,
            "hum": 0.24854575163398676,
            "windspeed": 0.18021650326797378,
            "users": 0.3483888008504226,
            "month": 0.9151846785225718,
            "hour": 0.0,
            "weekday": 0.055485194828118825,
        },
        "empirical_threshold_chi2_": {
            "holiday": np.nan,
            "workingday": np.nan,
            "weather": 128.71740772446282,
            "temp": np.nan,
            "atemp": np.nan,
            "hum": np.nan,
            "windspeed": np.nan,
            "users": np.nan,
            "month": np.nan,
            "hour": np.nan,
            "weekday": np.nan,
        },
        "empirical_threshold_js_": {
            "holiday": 0.12263338812190272,
            "workingday": 0.05536149547827185,
            "weather": 0.14695782935496615,
            "temp": 0.6845879616136854,
            "atemp": 0.6797815908883446,
            "hum": 0.28687940327125355,
            "windspeed": 0.1881001299842457,
            "users": 0.4033478565343228,
            "month": 0.892720109868587,
            "hour": 0.0,
            "weekday": 0.04868637565402696,
        },
        "ref_ranges_": {
            "holiday": (0.0, 1.0),
            "workingday": (0.0, 1.0),
            "temp": (0.8200000000000001, 41.0),
            "atemp": (0.0, 50.0),
            "hum": (0.0, 100.0),
            "windspeed": (0.0, 56.9969),
            "users": (1.0, 977.0),
            "month": (1, 12),
            "hour": (0, 23),
            "weekday": (0, 6),
        },
        "ref_categories_": {"weather": ["clear", "mist", "rain"]},
        "n_chunks_reference_data_": 24,
        "detectors_": {},
        "threshold": 0.95,
        "chunk_size": "ME",
    }

    for attr, expected_value in expected_results.items():
        value = getattr(detector, attr)
        if isinstance(value, dict):
            for key in expected_value.keys():
                if isinstance(expected_value[key], float) and np.isnan(expected_value[key]):
                    assert np.isnan(value[key])
                elif isinstance(expected_value[key], list):
                    assert value[key] == expected_value[key]
                else:
                    np.testing.assert_almost_equal(value[key], expected_value[key], decimal=5)
        elif isinstance(value, list):
            assert value == expected_value
        elif isinstance(value, float):
            np.testing.assert_almost_equal(value, expected_value, decimal=5)
        else:
            assert value == expected_value


def test_fit_stored_attributes_multiseries():
    """
    Test that the attributes of PopulationDriftDetector are correctly stored after calling fit
    when data contains multiple series.
    """

    detector = PopulationDriftDetector(chunk_size="ME", threshold=0.95)
    detector.fit(data_multiseries)

    expected_results = {
        "ref_features_": [
            "holiday",
            "workingday",
            "weather",
            "temp",
            "atemp",
            "hum",
            "windspeed",
            "users",
            "month",
            "hour",
            "weekday",
        ],
        "is_fitted_": True,
        "empirical_threshold_ks_": {
            "holiday": 0.03659032405160117,
            "workingday": 0.05896694761925775,
            "weather": np.nan,
            "temp": 0.6644120588970772,
            "atemp": 0.6419687642498859,
            "hum": 0.24854575163398676,
            "windspeed": 0.18021650326797378,
            "users": 0.3483888008504226,
            "month": 0.9151846785225718,
            "hour": 0.0,
            "weekday": 0.055485194828118825,
        },
        "empirical_threshold_chi2_": {
            "holiday": np.nan,
            "workingday": np.nan,
            "weather": 128.71740772446282,
            "temp": np.nan,
            "atemp": np.nan,
            "hum": np.nan,
            "windspeed": np.nan,
            "users": np.nan,
            "month": np.nan,
            "hour": np.nan,
            "weekday": np.nan,
        },
        "empirical_threshold_js_": {
            "holiday": 0.12263338812190272,
            "workingday": 0.05536149547827185,
            "weather": 0.14695782935496615,
            "temp": 0.6845879616136854,
            "atemp": 0.6797815908883446,
            "hum": 0.28687940327125355,
            "windspeed": 0.1881001299842457,
            "users": 0.4033478565343228,
            "month": 0.892720109868587,
            "hour": 0.0,
            "weekday": 0.04868637565402696,
        },
        "ref_ranges_": {
            "holiday": (0.0, 1.0),
            "workingday": (0.0, 1.0),
            "temp": (0.8200000000000001, 41.0),
            "atemp": (0.0, 50.0),
            "hum": (0.0, 100.0),
            "windspeed": (0.0, 56.9969),
            "users": (1.0, 977.0),
            "month": (1, 12),
            "hour": (0, 23),
            "weekday": (0, 6),
        },
        "ref_categories_": {"weather": ["clear", "mist", "rain"]},
        "n_chunks_reference_data_": 24,
        "detectors_": {},
        "threshold": 0.95,
        "chunk_size": "ME",
    }

    for series_id in ['series_1', 'series_2', 'series_3']:
        detector_series = detector.detectors_[series_id]
        for attr, expected_value in expected_results.items():
            value = getattr(detector_series, attr)
            if isinstance(value, dict):
                for key in expected_value.keys():
                    if isinstance(expected_value[key], float) and np.isnan(expected_value[key]):
                        assert np.isnan(value[key])
                    elif isinstance(expected_value[key], list):
                        assert value[key] == expected_value[key]
                    else:
                        np.testing.assert_almost_equal(value[key], expected_value[key], decimal=5)
            elif isinstance(value, list):
                assert value == expected_value
            elif isinstance(value, float):
                np.testing.assert_almost_equal(value, expected_value, decimal=5)
            else:
                assert value == expected_value



# Unit test _extract_data_folds_multiseries
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.model_selection._utils import _extract_data_folds_multiseries

# Fixtures
series_wide_range = pd.DataFrame({
    'l1': np.arange(50, dtype=float),
    'l2': np.arange(50, 100, dtype=float),
    'l3': np.arange(100, 150, dtype=float)
})
exog_wide_range = pd.DataFrame({
    'exog_1': np.arange(1000, 1050, dtype=float),
    'exog_2': np.arange(1050, 1100, dtype=float),
    'exog_3': np.arange(1100, 1150, dtype=float)
})


@pytest.mark.parametrize("dropna_last_window", 
                         [True, False], 
                         ids=lambda dropna: f'dropna_last_window: {dropna}')
def test_extract_data_folds_multiseries_series_DataFrame_exog_None_RangeIndex(dropna_last_window):
    """
    Test _extract_data_folds_multiseries with series DataFrame and exog None.
    """

    # Train, last_window, test_no_gap
    folds = [
        [0, [0, 30], [25, 30], [30, 37], True], 
        [1, [0, 35], [30, 35], [35, 42], True]
    ]
    span_index = pd.RangeIndex(start=0, stop=50, step=1)
    window_size = 5

    data_folds = list(
        _extract_data_folds_multiseries(
            series             = series_wide_range, 
            folds              = folds, 
            span_index         = span_index, 
            window_size        = window_size, 
            exog               = None, 
            dropna_last_window = dropna_last_window, 
            externally_fitted  = False
        )
    )

    expected_data_folds = [
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 30, dtype=float),
                        'l2': np.arange(50, 80, dtype=float),
                        'l3': np.arange(100, 130, dtype=float)
                },
                index = pd.RangeIndex(start=0, stop=30, step=1)
            ), 
            pd.DataFrame(
                data = {'l1': np.arange(25, 30, dtype=float),
                        'l2': np.arange(75, 80, dtype=float),
                        'l3': np.arange(125, 130, dtype=float)
                },
                index = pd.RangeIndex(start=25, stop=30, step=1)
            ),
            ['l1', 'l2', 'l3'],
            None, 
            None, 
            folds[0]
        ),
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 35, dtype=float),
                        'l2': np.arange(50, 85, dtype=float),
                        'l3': np.arange(100, 135, dtype=float)
                },
                index = pd.RangeIndex(start=0, stop=35, step=1)
            ), 
            pd.DataFrame(
                data = {'l1': np.arange(30, 35, dtype=float),
                        'l2': np.arange(80, 85, dtype=float),
                        'l3': np.arange(130, 135, dtype=float)
                },
                index = pd.RangeIndex(start=30, stop=35, step=1)
            ),
            ['l1', 'l2', 'l3'],
            None,
            None,
            folds[1]
        )
    ]

    for i, data_fold in enumerate(data_folds):

        assert isinstance(data_fold, tuple)
        assert len(data_fold) == 6
        
        pd.testing.assert_frame_equal(data_fold[0], expected_data_folds[i][0])
        pd.testing.assert_frame_equal(data_fold[1], expected_data_folds[i][1])
        assert data_fold[2] == expected_data_folds[i][2]
        assert data_fold[3] is None
        assert data_fold[4] is None
        assert data_fold[5] == expected_data_folds[i][5]


@pytest.mark.parametrize("dropna_last_window", 
                         [True, False], 
                         ids=lambda dropna: f'dropna_last_window: {dropna}')
def test_extract_data_folds_multiseries_series_DataFrame_with_NaN_exog_RangeIndex(dropna_last_window):
    """
    Test _extract_data_folds_multiseries with series DataFrame and exog None with
    RangeIndex.
    """
    series_nan = series_wide_range.copy()
    series_nan.loc[:28, 'l2'] = np.nan
    series_nan['l3'] = np.nan

    # Train, last_window, test_no_gap
    folds = [
        [0, [0, 30], [25, 30], [30, 37], True], 
        [1, [0, 35], [30, 35], [35, 42], True]
    ]
    span_index = pd.RangeIndex(start=0, stop=50, step=1)
    window_size = 5

    data_folds = list(
        _extract_data_folds_multiseries(
            series             = series_nan, 
            folds              = folds, 
            span_index         = span_index, 
            window_size        = window_size, 
            exog               = exog_wide_range, 
            dropna_last_window = dropna_last_window, 
            externally_fitted  = False
        )
    )

    expected_data_folds = [
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 30, dtype=float)
                },
                index = pd.RangeIndex(start=0, stop=30, step=1)
            ),
            pd.DataFrame(
                data = {'l1': np.arange(25, 30, dtype=float)
                },
                index = pd.RangeIndex(start=25, stop=30, step=1)
            ),
            ['l1'],
            pd.DataFrame(
                data = {'exog_1': np.arange(1000, 1030, dtype=float),
                        'exog_2': np.arange(1050, 1080, dtype=float),
                        'exog_3': np.arange(1100, 1130, dtype=float)
                },
                index = pd.RangeIndex(start=0, stop=30, step=1)
            ),
            pd.DataFrame(
                data = {'exog_1': np.arange(1030, 1037, dtype=float),
                        'exog_2': np.arange(1080, 1087, dtype=float),
                        'exog_3': np.arange(1130, 1137, dtype=float)
                },
                index = pd.RangeIndex(start=30, stop=37, step=1)
            ),
            folds[0]
        ),
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 35, dtype=float),
                        'l2': [np.nan] * 29 + list(range(79, 85))
                },
                index = pd.RangeIndex(start=0, stop=35, step=1)
            ).astype({'l2': float}), 
            pd.DataFrame(
                data = {'l1': np.arange(30, 35, dtype=float),
                        'l2': np.arange(80, 85, dtype=float)
                },
                index = pd.RangeIndex(start=30, stop=35, step=1)
            ),
            ['l1', 'l2'],
            pd.DataFrame(
                data = {'exog_1': np.arange(1000, 1035, dtype=float),
                        'exog_2': np.arange(1050, 1085, dtype=float),
                        'exog_3': np.arange(1100, 1135, dtype=float)
                },
                index = pd.RangeIndex(start=0, stop=35, step=1)
            ),
            pd.DataFrame(
                data = {'exog_1': np.arange(1035, 1042, dtype=float),
                        'exog_2': np.arange(1085, 1092, dtype=float),
                        'exog_3': np.arange(1135, 1142, dtype=float)
                },
                index = pd.RangeIndex(start=35, stop=42, step=1)
            ),
            folds[1]
        )
    ]

    for i, data_fold in enumerate(data_folds):

        assert isinstance(data_fold, tuple)
        assert len(data_fold) == 6
        
        pd.testing.assert_frame_equal(data_fold[0], expected_data_folds[i][0])
        pd.testing.assert_frame_equal(data_fold[1], expected_data_folds[i][1])
        assert data_fold[2] == expected_data_folds[i][2]
        pd.testing.assert_frame_equal(data_fold[3], expected_data_folds[i][3])
        pd.testing.assert_frame_equal(data_fold[4], expected_data_folds[i][4])
        assert data_fold[5] == expected_data_folds[i][5]


def test_extract_data_folds_multiseries_series_DataFrame_with_NaN_exog_DatetimeIndex():
    """
    Test _extract_data_folds_multiseries with series DataFrame with NaNs and exog 
    with DatetimeIndex.
    """
    series_nan = series_wide_range.copy()
    series_nan.loc[28, 'l2'] = np.nan  # 29th value of 'l2' is NaN (affected by window_size=5)
    series_nan.loc[:15, 'l3'] = np.nan  # First 16 values of 'l3' are NaN

    series_nan.index = pd.date_range(start='2020-01-01', periods=50, freq='D')
    exog_datetime = exog_wide_range.copy()
    exog_datetime.index = pd.date_range(start='2020-01-01', periods=50, freq='D')

    # Train, last_window, test_no_gap
    folds = [
        [0, [0, 30], [25, 30], [30, 37], True], 
        [1, [0, 35], [30, 35], [35, 42], True]
    ]
    span_index = pd.date_range(start='2020-01-01', periods=50, freq='D')
    window_size = 5

    data_folds = list(
        _extract_data_folds_multiseries(
            series             = series_nan, 
            folds              = folds, 
            span_index         = span_index, 
            window_size        = window_size, 
            exog               = exog_datetime, 
            dropna_last_window = True, 
            externally_fitted  = False
        )
    )

    expected_data_folds = [
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 30, dtype=float),
                        'l2': list(range(50, 78)) + [np.nan] + [79],
                        'l3': [np.nan] * 16 + list(range(116, 130))
                },
                index = pd.date_range(start='2020-01-01', periods=30, freq='D')
            ).astype({'l2': float, 'l3': float}),
            pd.DataFrame(
                data = {'l1': np.arange(25, 30, dtype=float),
                        'l3': np.arange(125, 130, dtype=float)
                },
                index = pd.date_range(start='2020-01-26', periods=5, freq='D')
            ),
            ['l1', 'l3'],
            pd.DataFrame(
                data = {'exog_1': np.arange(1000, 1030, dtype=float),
                        'exog_2': np.arange(1050, 1080, dtype=float),
                        'exog_3': np.arange(1100, 1130, dtype=float)
                },
                index = pd.date_range(start='2020-01-01', periods=30, freq='D')
            ),
            pd.DataFrame(
                data = {'exog_1': np.arange(1030, 1037, dtype=float),
                        'exog_2': np.arange(1080, 1087, dtype=float),
                        'exog_3': np.arange(1130, 1137, dtype=float)
                },
                index = pd.date_range(start='2020-01-31', periods=7, freq='D')
            ),
            folds[0]
        ),
        (
            pd.DataFrame(
                data = {'l1': np.arange(0, 35, dtype=float),
                        'l2': list(range(50, 78)) + [np.nan] + list(range(79, 85)),
                        'l3': [np.nan] * 16 + list(range(116, 135))
                },
                index = pd.date_range(start='2020-01-01', periods=35, freq='D')
            ).astype({'l2': float}), 
            pd.DataFrame(
                data = {'l1': np.arange(30, 35, dtype=float),
                        'l2': np.arange(80, 85, dtype=float),
                        'l3': np.arange(130, 135, dtype=float)
                },
                index = pd.date_range(start='2020-01-31', periods=5, freq='D')
            ),
            ['l1', 'l2', 'l3'],
            pd.DataFrame(
                data = {'exog_1': np.arange(1000, 1035, dtype=float),
                        'exog_2': np.arange(1050, 1085, dtype=float),
                        'exog_3': np.arange(1100, 1135, dtype=float)
                },
                index = pd.date_range(start='2020-01-01', periods=35, freq='D')
            ),
            pd.DataFrame(
                data = {'exog_1': np.arange(1035, 1042, dtype=float),
                        'exog_2': np.arange(1085, 1092, dtype=float),
                        'exog_3': np.arange(1135, 1142, dtype=float)
                },
                index = pd.date_range(start='2020-02-05', periods=7, freq='D')
            ),
            folds[1]
        )
    ]

    for i, data_fold in enumerate(data_folds):

        assert isinstance(data_fold, tuple)
        assert len(data_fold) == 6
        
        pd.testing.assert_frame_equal(data_fold[0], expected_data_folds[i][0])
        pd.testing.assert_frame_equal(data_fold[1], expected_data_folds[i][1])
        assert data_fold[2] == expected_data_folds[i][2]
        pd.testing.assert_frame_equal(data_fold[3], expected_data_folds[i][3])
        pd.testing.assert_frame_equal(data_fold[4], expected_data_folds[i][4])
        assert data_fold[5] == expected_data_folds[i][5]


@pytest.mark.parametrize("dropna_last_window", 
                         [True, False], 
                         ids=lambda dropna: f'dropna_last_window: {dropna}')
def test_extract_data_folds_multiseries_series_dict_exog_dict_DatetimeIndex(dropna_last_window):
    """
    Test _extract_data_folds_multiseries with series and exog as dict with
    DatetimeIndex.
    """
    series_wide_dt = series_wide_range.copy()
    series_wide_dt.index = pd.date_range(start='2020-01-01', periods=50, freq='D')

    series_dict = {
        'l1': series_wide_dt['l1'],
        'l2': series_wide_dt['l2'].iloc[10:],
        'l3': series_wide_dt['l3'].iloc[:20]
    }

    exog_wide_dt = exog_wide_range.copy()
    exog_wide_dt.index = pd.date_range(start='2020-01-01', periods=50, freq='D')

    exog_dict = {
        'l1': exog_wide_dt,
        'l2': exog_wide_dt[['exog_1', 'exog_2']].iloc[15:],
        'l3': exog_wide_dt['exog_3'].iloc[30:]
    }

    # Train, last_window, test_no_gap
    folds = [
        [0, [0, 30], [25, 30], [30, 37], True], 
        [1, [0, 35], [30, 35], [35, 42], True]
    ]
    span_index = pd.date_range(start='2020-01-01', periods=50, freq='D')
    window_size = 5

    data_folds = list(
        _extract_data_folds_multiseries(
            series             = series_dict, 
            folds              = folds, 
            span_index         = span_index, 
            window_size        = window_size, 
            exog               = exog_dict, 
            dropna_last_window = dropna_last_window, 
            externally_fitted  = False
        )
    )

    expected_data_folds = [
        (
            {
                'l1': pd.Series(
                          data  = np.arange(0, 30, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=30, freq='D'),
                          name  = 'l1'
                      ),
                'l2': pd.Series(
                          data  = np.arange(60, 80, dtype=float), 
                          index = pd.date_range(start='2020-01-11', periods=20, freq='D'),
                          name  = 'l2'
                      ),
                'l3': pd.Series(
                          data  = np.arange(100, 120, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=20, freq='D'),
                          name  = 'l3'
                      )
            },
            pd.DataFrame(
                data = {'l1': np.arange(25, 30, dtype=float),
                        'l2': np.arange(75, 80, dtype=float)
                },
                index = pd.date_range(start='2020-01-26', periods=5, freq='D')
            ),
            ['l1', 'l2'],
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1000, 1030, dtype=float),
                                   'exog_2': np.arange(1050, 1080, dtype=float),
                                   'exog_3': np.arange(1100, 1130, dtype=float)}, 
                          index = pd.date_range(start='2020-01-01', periods=30, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1015, 1030, dtype=float),
                                   'exog_2': np.arange(1065, 1080, dtype=float)}, 
                          index = pd.date_range(start='2020-01-16', periods=15, freq='D')
                      )
            },
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1030, 1037, dtype=float),
                                   'exog_2': np.arange(1080, 1087, dtype=float),
                                   'exog_3': np.arange(1130, 1137, dtype=float)}, 
                          index = pd.date_range(start='2020-01-31', periods=7, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1030, 1037, dtype=float),
                                   'exog_2': np.arange(1080, 1087, dtype=float)}, 
                          index = pd.date_range(start='2020-01-31', periods=7, freq='D')
                      )
            },
            folds[0]
        ),
        (
            {
                'l1': pd.Series(
                          data  = np.arange(0, 35, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=35, freq='D'),
                          name  = 'l1'
                      ),
                'l2': pd.Series(
                          data  = np.arange(60, 85, dtype=float), 
                          index = pd.date_range(start='2020-01-11', periods=25, freq='D'),
                          name  = 'l2'
                      ),
                'l3': pd.Series(
                          data  = np.arange(100, 120, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=20, freq='D'),
                          name  = 'l3'
                      )
            }, 
            pd.DataFrame(
                data = {'l1': np.arange(30, 35, dtype=float),
                        'l2': np.arange(80, 85, dtype=float)
                },
                index = pd.date_range(start='2020-01-31', periods=5, freq='D')
            ),
            ['l1', 'l2'],
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1000, 1035, dtype=float),
                                   'exog_2': np.arange(1050, 1085, dtype=float),
                                   'exog_3': np.arange(1100, 1135, dtype=float)}, 
                          index = pd.date_range(start='2020-01-01', periods=35, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1015, 1035, dtype=float),
                                   'exog_2': np.arange(1065, 1085, dtype=float)}, 
                          index = pd.date_range(start='2020-01-16', periods=20, freq='D')
                      ),
                'l3': pd.Series(
                          data  = np.arange(1130, 1135, dtype=float), 
                          index = pd.date_range(start='2020-01-31', periods=5, freq='D'),
                          name  = 'exog_3'
                      )
            },
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1035, 1042, dtype=float),
                                   'exog_2': np.arange(1085, 1092, dtype=float),
                                   'exog_3': np.arange(1135, 1142, dtype=float)}, 
                          index = pd.date_range(start='2020-02-05', periods=7, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1035, 1042, dtype=float),
                                   'exog_2': np.arange(1085, 1092, dtype=float)}, 
                          index = pd.date_range(start='2020-02-05', periods=7, freq='D')
                      ),
                'l3': pd.Series(
                          data  = np.arange(1135, 1142, dtype=float), 
                          index = pd.date_range(start='2020-02-05', periods=7, freq='D'),
                          name  = 'exog_3'
                      )
            },
            folds[1]
        )
    ]

    for i, data_fold in enumerate(data_folds):

        assert isinstance(data_fold, tuple)
        assert len(data_fold) == 6
        
        for key in data_fold[0].keys():
            pd.testing.assert_series_equal(data_fold[0][key], expected_data_folds[i][0][key])
        pd.testing.assert_frame_equal(data_fold[1], expected_data_folds[i][1])

        assert data_fold[2] == expected_data_folds[i][2]

        for key in data_fold[3].keys():
            if isinstance(expected_data_folds[i][3][key], pd.DataFrame):
                pd.testing.assert_frame_equal(data_fold[3][key], expected_data_folds[i][3][key])
            else:
                pd.testing.assert_series_equal(data_fold[3][key], expected_data_folds[i][3][key])
        for key in data_fold[4].keys():
            if isinstance(expected_data_folds[i][4][key], pd.DataFrame):
                pd.testing.assert_frame_equal(data_fold[4][key], expected_data_folds[i][4][key])
            else:
                pd.testing.assert_series_equal(data_fold[4][key], expected_data_folds[i][4][key])

        assert data_fold[5] == expected_data_folds[i][5]


@pytest.mark.parametrize("dropna_last_window", 
                         [True, False], 
                         ids=lambda dropna: f'dropna_last_window: {dropna}')
def test_extract_data_folds_multiseries_series_dict_exog_dict_DatetimeIndex_with_None_in_exog(dropna_last_window):
    """
    Test _extract_data_folds_multiseries with series and exog as dict with
    DatetimeIndex. Exog contains None values.
    """
    series_wide_dt = series_wide_range.copy()
    series_wide_dt.index = pd.date_range(start='2020-01-01', periods=50, freq='D')

    series_dict = {
        'l1': series_wide_dt['l1'],
        'l2': series_wide_dt['l2'].iloc[10:],
        'l3': series_wide_dt['l3'].iloc[:20]
    }

    exog_wide_dt = exog_wide_range.copy()
    exog_wide_dt.index = pd.date_range(start='2020-01-01', periods=50, freq='D')

    exog_dict = {
        'l1': exog_wide_dt,
        'l2': exog_wide_dt[['exog_1', 'exog_2']].iloc[15:],
        'l3': None
    }

    # Train, last_window, test_no_gap
    folds = [
        [0, [0, 30], [25, 30], [30, 37], True], 
        [1, [0, 35], [30, 35], [35, 42], True]
    ]
    span_index = pd.date_range(start='2020-01-01', periods=50, freq='D')
    window_size = 5

    data_folds = list(
        _extract_data_folds_multiseries(
            series             = series_dict, 
            folds              = folds, 
            span_index         = span_index, 
            window_size        = window_size, 
            exog               = exog_dict, 
            dropna_last_window = dropna_last_window, 
            externally_fitted  = False
        )
    )

    expected_data_folds = [
        (
            {
                'l1': pd.Series(
                          data  = np.arange(0, 30, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=30, freq='D'),
                          name  = 'l1'
                      ),
                'l2': pd.Series(
                          data  = np.arange(60, 80, dtype=float), 
                          index = pd.date_range(start='2020-01-11', periods=20, freq='D'),
                          name  = 'l2'
                      ),
                'l3': pd.Series(
                          data  = np.arange(100, 120, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=20, freq='D'),
                          name  = 'l3'
                      )
            },
            pd.DataFrame(
                data = {'l1': np.arange(25, 30, dtype=float),
                        'l2': np.arange(75, 80, dtype=float)
                },
                index = pd.date_range(start='2020-01-26', periods=5, freq='D')
            ),
            ['l1', 'l2'],
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1000, 1030, dtype=float),
                                   'exog_2': np.arange(1050, 1080, dtype=float),
                                   'exog_3': np.arange(1100, 1130, dtype=float)}, 
                          index = pd.date_range(start='2020-01-01', periods=30, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1015, 1030, dtype=float),
                                   'exog_2': np.arange(1065, 1080, dtype=float)}, 
                          index = pd.date_range(start='2020-01-16', periods=15, freq='D')
                      ),
                'l3': None
            },
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1030, 1037, dtype=float),
                                   'exog_2': np.arange(1080, 1087, dtype=float),
                                   'exog_3': np.arange(1130, 1137, dtype=float)}, 
                          index = pd.date_range(start='2020-01-31', periods=7, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1030, 1037, dtype=float),
                                   'exog_2': np.arange(1080, 1087, dtype=float)}, 
                          index = pd.date_range(start='2020-01-31', periods=7, freq='D')
                      ),
                'l3': None
            },
            folds[0]
        ),
        (
            {
                'l1': pd.Series(
                          data  = np.arange(0, 35, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=35, freq='D'),
                          name  = 'l1'
                      ),
                'l2': pd.Series(
                          data  = np.arange(60, 85, dtype=float), 
                          index = pd.date_range(start='2020-01-11', periods=25, freq='D'),
                          name  = 'l2'
                      ),
                'l3': pd.Series(
                          data  = np.arange(100, 120, dtype=float), 
                          index = pd.date_range(start='2020-01-01', periods=20, freq='D'),
                          name  = 'l3'
                      )
            }, 
            pd.DataFrame(
                data = {'l1': np.arange(30, 35, dtype=float),
                        'l2': np.arange(80, 85, dtype=float)
                },
                index = pd.date_range(start='2020-01-31', periods=5, freq='D')
            ),
            ['l1', 'l2'],
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1000, 1035, dtype=float),
                                   'exog_2': np.arange(1050, 1085, dtype=float),
                                   'exog_3': np.arange(1100, 1135, dtype=float)}, 
                          index = pd.date_range(start='2020-01-01', periods=35, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1015, 1035, dtype=float),
                                   'exog_2': np.arange(1065, 1085, dtype=float)}, 
                          index = pd.date_range(start='2020-01-16', periods=20, freq='D')
                      ),
                'l3': None
            },
            {
                'l1': pd.DataFrame(
                          data  = {'exog_1': np.arange(1035, 1042, dtype=float),
                                   'exog_2': np.arange(1085, 1092, dtype=float),
                                   'exog_3': np.arange(1135, 1142, dtype=float)}, 
                          index = pd.date_range(start='2020-02-05', periods=7, freq='D')
                      ),
                'l2': pd.DataFrame(
                          data  = {'exog_1': np.arange(1035, 1042, dtype=float),
                                   'exog_2': np.arange(1085, 1092, dtype=float)}, 
                          index = pd.date_range(start='2020-02-05', periods=7, freq='D')
                      ),
                'l3': None
            },
            folds[1]
        )
    ]

    for i, data_fold in enumerate(data_folds):

        assert isinstance(data_fold, tuple)
        assert len(data_fold) == 6
        
        for key in data_fold[0].keys():
            pd.testing.assert_series_equal(data_fold[0][key], expected_data_folds[i][0][key])
        pd.testing.assert_frame_equal(data_fold[1], expected_data_folds[i][1])

        assert data_fold[2] == expected_data_folds[i][2]

        for key in data_fold[3].keys():
            if expected_data_folds[i][3][key] is None:
                assert data_fold[3][key] is None
            else:
                if isinstance(expected_data_folds[i][3][key], pd.DataFrame):
                    pd.testing.assert_frame_equal(data_fold[3][key], expected_data_folds[i][3][key])
                else:
                    pd.testing.assert_series_equal(data_fold[3][key], expected_data_folds[i][3][key])
        for key in data_fold[4].keys():
            if expected_data_folds[i][4][key] is None:
                assert data_fold[4][key] is None
            else:
                if isinstance(expected_data_folds[i][4][key], pd.DataFrame):
                    pd.testing.assert_frame_equal(data_fold[4][key], expected_data_folds[i][4][key])
                else:
                    pd.testing.assert_series_equal(data_fold[4][key], expected_data_folds[i][4][key])

        assert data_fold[5] == expected_data_folds[i][5]



# Unit test OneStepAheadFold
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.model_selection._split import OneStepAheadFold


def test_OneStepAheadFold_split_raise_error_when_X_is_not_series_dataframe_or_dict():
    """
    Test that ValueError is raised when X is not a pd.Series, pd.DataFrame or dict.
    """
    X = np.arange(100)
    cv = OneStepAheadFold(initial_train_size=70)
    
    err_msg = re.escape(
        f"X must be a pandas Series, DataFrame, Index or a dictionary. "
        f"Got {type(X)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        cv.split(X=X)


@pytest.mark.parametrize("invalid_date", [
    "2021-12-31",  # Before the first date in the index
    "2022-04-11",  # After the last date in the index
])
def test_OneStepAhead_split_invalid_initial_train_size_date(invalid_date):
    """
    Test that ValueError is raised when initial_train_size date is outside the index range.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start="2022-01-01", periods=100, freq="D")
    cv = OneStepAheadFold(
        initial_train_size=invalid_date,
        window_size=3,
        differentiation=None
    )
    
    err_msg = re.escape(
        "If `initial_train_size` is a date, it must be greater than "
        "the first date in the index and less than the last date."
    )
    with pytest.raises(ValueError, match=err_msg):
        cv.split(X=y)


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [0, [range(0, 70)], [range(70, 100)], True]),
                          (False, [0, [0, 70], [70, 100], True])], 
                         ids = lambda argument: f'{argument}')
def test_OneStepAhead_split_initial_train_size_and_window_size(capfd, return_all_indexes, expected):
    """
    Test OneStepAhead splits when initial_train_size and window_size are provided.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = OneStepAheadFold(
            initial_train_size = 70,
            window_size        = 3,
            differentiation    = None,
            return_all_indexes = return_all_indexes,
        )
    folds = cv.split(X=y)
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations in train: 70\n"
        "Number of observations in test: 30\n"
        "Training : 2022-01-01 00:00:00 -- 2022-03-11 00:00:00 (n=70)\n"
        "Test     : 2022-03-12 00:00:00 -- 2022-04-10 00:00:00 (n=30)\n\n"
    )

    assert out == expected_out
    assert folds == expected


def test_OneStepAhead_split_initial_train_size_and_window_size_differentiation_is_1(capfd):
    """
    Test OneStepAhead splits when initial_train_size and window_size are provided, and
    differentiation is 1.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = OneStepAheadFold(
            initial_train_size = 70,
            window_size        = 3,
            differentiation    = 1,
            return_all_indexes = False,
        )
    folds = cv.split(X=y)
    out, _ = capfd.readouterr()
    expected_folds = [0, [0, 70], [70, 100], True]
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations in train: 69\n"
        "    First 1 observation/s in training set are used for differentiation\n"
        "Number of observations in test: 30\n"
        "Training : 2022-01-02 00:00:00 -- 2022-03-11 00:00:00 (n=69)\n"
        "Test     : 2022-03-12 00:00:00 -- 2022-04-10 00:00:00 (n=30)\n\n"
    )

    assert out == expected_out
    assert folds == expected_folds


def test_OneStepAhead_split_initial_train_size_window_size_return_all_indexes_true_as_pandas_true(capfd):
    """
    Test OneStepAhead splits when initial_train_size and window_size are provided, output as
    pandas DataFrame.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = OneStepAheadFold(
            initial_train_size = 70,
            window_size        = 3,
            differentiation    = None,
            return_all_indexes = True,
        )
    folds = cv.split(X=y, as_pandas=True)
    out, _ = capfd.readouterr()
    expected_folds = pd.DataFrame(
        {
            "fold": {0: 0},
            "train_index": {0: [range(0, 70)]},
            "test_index": {0: [range(70, 100)]},
            "fit_forecaster": {0: True},
        }
    )
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations in train: 70\n"
        "Number of observations in test: 30\n"
        "Training : 2022-01-01 00:00:00 -- 2022-03-11 00:00:00 (n=70)\n"
        "Test     : 2022-03-12 00:00:00 -- 2022-04-10 00:00:00 (n=30)\n\n"
    )

    assert out == expected_out
    pd.testing.assert_frame_equal(folds, expected_folds)


def test_OneStepAhead_split_initial_train_size_window_size_return_all_indexes_false_as_pandas_true(capfd):
    """
    Test OneStepAhead splits when initial_train_size and window_size are provided, output as
    pandas DataFrame.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    cv = OneStepAheadFold(
            initial_train_size = 70,
            window_size        = 3,
            differentiation    = None,
            return_all_indexes = False,
        )
    folds = cv.split(X=y, as_pandas=True)
    out, _ = capfd.readouterr()
    expected_folds = pd.DataFrame(
        {
            'fold': [0],
            'train_start': [0],
            'train_end': [70],
            'test_start': [70],
            'test_end': [100],
            'fit_forecaster': [True]
        }
    )
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations in train: 70\n"
        "Number of observations in test: 30\n"
        "Training : 2022-01-01 00:00:00 -- 2022-03-11 00:00:00 (n=70)\n"
        "Test     : 2022-03-12 00:00:00 -- 2022-04-10 00:00:00 (n=30)\n\n"
    )

    assert out == expected_out
    pd.testing.assert_frame_equal(folds, expected_folds)


@pytest.mark.parametrize(
    "initial_train_size, expected",
    [
        (70, [0, [0, 70], [70, 100], True]),
        ("2022-03-11", [0, [0, 70], [70, 100], True]),
        ("2022-03-11 00:00:00", [0, [0, 70], [70, 100], True]),
        (pd.to_datetime("2022-03-11"), [0, [0, 70], [70, 100], True])
    ],
    ids=lambda x: f'initial_train_size={x}',
)
def test_OneStepAhead_split_int_and_date_initial_train_size(capfd, initial_train_size, expected):
    """
    Test OneStepAhead splits with different types for initial_train_size.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start="2022-01-01", periods=100, freq="D")
    cv = OneStepAheadFold(
        initial_train_size=initial_train_size,
        window_size=3,
        differentiation=None
    )
    folds = cv.split(X=y)
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of folds\n"
        "--------------------\n"
        "Number of observations in train: 70\n"
        "Number of observations in test: 30\n"
        "Training : 2022-01-01 00:00:00 -- 2022-03-11 00:00:00 (n=70)\n"
        "Test     : 2022-03-12 00:00:00 -- 2022-04-10 00:00:00 (n=30)\n\n"
    )

    assert out == expected_out
    assert folds == expected
